/*****************************************************************
OpenCL parallel integrator based on Runge-Kutta schemes.

Copyright 2014 Marcos Rodriguez, Fernando Blesa and Roberto Barrio
http://cody.unizar.es/software/
Contact: <marcos@unizar.es>

This code is released under GPL license. See the web site above for citation instructions

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*****************************************************************/


int status;
	
size_t devInfo, devInfos[3];
char dname[500];

// Get the first available platform
// Example: AMD Accelerated Parallel Processing
cl_platform_id platform[1];
clGetPlatformIDs(1,			// number of platforms to add to list
		platform, 		// list of platforms found
		NULL);			// number of platforms available

// Get the first SELECT CPU OR GPU device the platform provides
cl_device_id device;
#ifdef CPU
clGetDeviceIDs(platform[0], CL_DEVICE_TYPE_CPU, 
		1, 			// number of devices to add
		&device, 		// list of devices
		NULL);			// number of devices available
#endif
#ifdef GPU
clGetDeviceIDs(platform[0], CL_DEVICE_TYPE_GPU, 
		1, 			// number of devices to add
		&device, 		// list of devices
		NULL);			// number of devices available
#endif

clGetPlatformInfo (platform[0], CL_PLATFORM_NAME,500,dname,NULL);
printf ("CL_PLATFORM_NAME = %s\n", dname);
clGetDeviceInfo (device, CL_DEVICE_NAME, 500, dname,NULL);
printf ("\tDevice name = %s\n", dname);
clGetDeviceInfo (device, CL_DEVICE_MAX_COMPUTE_UNITS, 
		sizeof (size_t), &devInfo, NULL);
printf ("\tMax compute Units = %i\n", (int) devInfo);
clGetDeviceInfo (device, CL_DEVICE_MAX_WORK_GROUP_SIZE, 
		sizeof (size_t), &devInfo, NULL);
printf ("\tMax work group size = %i\n", (int) devInfo);
clGetDeviceInfo (device, CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,
		sizeof (size_t), &devInfo, NULL);
printf ("\tMax number of dimensions = %i\n", (int) devInfo);
clGetDeviceInfo (device, CL_DEVICE_MAX_WORK_ITEM_SIZES,
		3*sizeof (size_t), devInfos, NULL);
printf ("\tMax work group sizes = %i, %i, %i\n\n", 
	(int) devInfos[0], (int) devInfos[1], (int) devInfos[2]);


clGetDeviceInfo (device, CL_DEVICE_NAME, 500, dname,NULL);


cl_context context = clCreateContext (
		0,			// optional (context properties)
		1,			// number of devices
		&device,		// pointer to device list
		NULL, NULL, 		// optional (callback func for reporting errors)
		&status);		// no error code returned
if (status != CL_SUCCESS) {
	printf ("create context error = %d\n", status);
	return 1;
}
cl_command_queue queue = clCreateCommandQueue (
		context,		// valid context
		device,			// device associated with context
		CL_QUEUE_PROFILING_ENABLE,	// optional (command queue properties)
		&status);		// no error code returned
if (status != CL_SUCCESS) {
	printf ("create queue error = %d\n", status);
	return 1;
}


