/*****************************************************************
OpenCL parallel integrator based on Runge-Kutta schemes.

Copyright 2014 Marcos Rodriguez, Fernando Blesa and Roberto Barrio
http://cody.unizar.es/software/
Contact: <marcos@unizar.es>

This code is released under GPL license. See the web site above for citation instructions

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*****************************************************************/

#ifdef LOWORDER
unsigned char rkstep (real_t *rk1, real_t *rkErr, real_t *x, real_t step) {
	
	unsigned char j;
	real_t rk2[NVAR];

	for (j=0; j<NVAR; j++) {
			rk1[j] = x[j];
			rk2[j] = x[j];
	}
	// STAGE 1
	f(rk2, rk2);
	for (j=0; j<NVAR; j++) {
			rkErr[j] = (B1-_B1)*rk2[j]*step;
			rk1[j] = rk1[j] + A21*rk2[j]*step;
			rk2[j] = rk1[j] + (B1-A21)*rk2[j]*step;
	}
	// STAGE 2
	f(rk1, rk1);
	for (j=0; j<NVAR; j++) {
		rkErr[j] = rkErr[j] + (B2-_B2)*rk1[j]*step;
		rk2[j] = rk2[j] + A32*rk1[j]*step;
		rk1[j] = rk2[j] + (B2-A32)*rk1[j]*step;
	}
	// STAGE 3
	f(rk2, rk2);	
	for (j=0; j<NVAR; j++) {
			rkErr[j] = rkErr[j] + (B3-_B3)*rk2[j]*step;
			rk1[j] = rk1[j] + A43*rk2[j]*step;
			rk2[j] = rk1[j] + (B3-A43)*rk2[j]*step;
	}
	// STAGE 4
	f(rk1, rk1);
	for (j=0; j<NVAR; j++){
		rkErr[j] = rkErr[j] + (B4-_B4)*rk1[j]*step;
		rk2[j] = rk2[j] + A54*rk1[j]*step;
		rk1[j] = rk2[j] + (B4-A54)*rk1[j]*step;
	}
	// STAGE 5
	f(rk2, rk2);	
	for (j=0; j<NVAR; j++){
		rkErr[j] = rkErr[j] + (B5-_B5)*rk2[j]*step;
		rk1[j] = rk1[j] + B5*rk2[j]*step;
	}

	// ESTIMATION OF LOCAL ERROR
	rkErr[0] = fabs (rkErr[0]);
	for (j=1; j<NVAR; j++)
		rkErr[0] = rkErr[0] + fabs(rkErr[j]);

	// RETURN ACCEPTANCE (1) OR REJECTION (0) OF step
	if (rkErr[0] < TOL) return 1;
	else return 0;
}
#endif



#ifdef HIGHORDER
unsigned char rkstep (real_t *rk1, real_t *rkErr, real_t *x, real_t step) {

	unsigned char j;
	real_t rk2[NVAR];

	for (j=0; j<NVAR; j++) {
			rk1[j] = x[j];
			rk2[j] = x[j];
	}
	// STAGE 1
	f(rk2, rk2);
	for (j=0; j<NVAR; j++) {
			rkErr[j] = (B1-_B1)*rk2[j]*step;
			rk1[j] = rk1[j] + A21*rk2[j]*step;
			rk2[j] = rk1[j] + (B1-A21)*rk2[j]*step;
	}
	// STAGE 2
	f(rk1, rk1);
	for (j=0; j<NVAR; j++) {
		rkErr[j] = rkErr[j] + (B2-_B2)*rk1[j]*step;
		rk2[j] = rk2[j] + A32*rk1[j]*step;
		rk1[j] = rk2[j] + (B2-A32)*rk1[j]*step;
	}
	// STAGE 3
	f(rk2, rk2);	
	for (j=0; j<NVAR; j++) {
			rkErr[j] = rkErr[j] + (B3-_B3)*rk2[j]*step;
			rk1[j] = rk1[j] + A43*rk2[j]*step;
			rk2[j] = rk1[j] + (B3-A43)*rk2[j]*step;
	}
	// STAGE 4
	f(rk1, rk1);
	for (j=0; j<NVAR; j++) {
		rkErr[j] = rkErr[j] + (B4-_B4)*rk1[j]*step;
		rk2[j] = rk2[j] + A54*rk1[j]*step;
		rk1[j] = rk2[j] + (B4-A54)*rk1[j]*step;
	}
	// STAGE 5
	f(rk2, rk2);	
	for (j=0; j<NVAR; j++) {
		rkErr[j] = rkErr[j] + (B5-_B5)*rk2[j]*step;
		rk1[j] = rk1[j] + A65*rk2[j]*step;
		rk2[j] = rk1[j] + (B5-A65)*rk2[j]*step;
	}
	// STAGE 6
	f(rk1, rk1);
	for (j=0; j<NVAR; j++) {
		rkErr[j] = rkErr[j] + (B6-_B6)*rk1[j]*step;
		rk2[j] = rk2[j] + A76*rk1[j]*step;
		rk1[j] = rk2[j] + (B6-A76)*rk1[j]*step;
	}
	// STAGE 7
	f(rk2, rk2);
	for (j=0; j<NVAR; j++) {
		rkErr[j] = rkErr[j] + (B7-_B7)*rk2[j]*step;
		rk1[j] = rk1[j] + A87*rk2[j]*step;
		rk2[j] = rk1[j] + (B7-A87)*rk2[j]*step;
	}
	// STAGE 8
	f(rk1, rk1);
	for (j=0; j<NVAR; j++) {
		rkErr[j] = rkErr[j] + (B8-_B8)*rk1[j]*step;
		rk2[j] = rk2[j] + A98*rk1[j]*step;
		rk1[j] = rk2[j] + (B8-A98)*rk1[j]*step;
	}
	// STAGE 9
	f(rk2, rk2);
	for (j=0; j<NVAR; j++){
		rkErr[j] = rkErr[j] + (B9-_B9)*rk2[j]*step;
		rk1[j] =rk1[j] + B9*rk2[j]*step;
	}


	// ESTIMATION OF LOCAL ERROR
	rkErr[0] = fabs (rkErr[0]);
	for (j=1; j<NVAR; j++)
		rkErr[0] = rkErr[0] + fabs(rkErr[j]);

	// RETURN ACCEPTANCE (1) OR REJECTION (0) OF step
	if (rkErr[0] < TOL) return 1;
	else return 0;
}

#endif



#ifdef DOPRI
unsigned char rkstep (real_t *xNew, real_t *rkErr, real_t *x, real_t step) {

	unsigned char j;
	real_t k1[NVAR], k2[NVAR], k3[NVAR], k4[NVAR], k5[NVAR], k6[NVAR], k7[NVAR];


	// STAGE 1
	for (j=0; j<NVAR; j++) {
		k1[j] = x[j];
	}
	f(k1, k1);
	for (j=0; j<NVAR; j++) {
		k1[j] = step*k1[j];
	}
	// STAGE 2
	for (j=0; j<NVAR; j++) {
		k2[j] = x[j] + A21*k1[j];
	}
	f(k2, k2);
	for (j=0; j<NVAR; j++) {
		k2[j] = step*k2[j];
	}	
	// STAGE 3
	for (j=0; j<NVAR; j++) {
		k3[j] = x[j] + A31*k1[j] + A32*k2[j];
	}
	f(k3, k3);
	for (j=0; j<NVAR; j++) {
		k3[j] = step*k3[j];
	}
	// STAGE 4
	for (j=0; j<NVAR; j++) {
		k4[j] = x[j] + A41*k1[j] + A42*k2[j] + A43*k3[j];
	}
	f(k4, k4);
	for (j=0; j<NVAR; j++) {
		k4[j] = step*k4[j];
	}
	// STAGE 5
	for (j=0; j<NVAR; j++) {
		k5[j] = x[j] + A51*k1[j] + A52*k2[j] + A53*k3[j] + A54*k4[j];
	}
	f(k5, k5);
	for (j=0; j<NVAR; j++) {
		k5[j] = step*k5[j];
	}
	// STAGE 6
	for (j=0; j<NVAR; j++) {
		k6[j] = x[j] + A61*k1[j] + A62*k2[j] + A63*k3[j] + A64*k4[j] + A65*k5[j];
	}
	f(k6, k6);
	for (j=0; j<NVAR; j++) {
		k6[j] = step*k6[j];
	}
	// STAGE 7
	for (j=0; j<NVAR; j++) {
		k7[j] = x[j] + A71*k1[j] + A72*k2[j] + A73*k3[j] + A74*k4[j] + A75*k5[j] + A76*k6[j];
	}
	f(k7, k7);
	for (j=0; j<NVAR; j++) {
		k7[j] = step*k7[j];
	}


	// EVAL AND ESTIMATE ERROR
	for (j=0; j<NVAR; j++) {
		rkErr[j] = (B1-_B1)*k1[j] + (B2-_B2)*k2[j] + (B3-_B3)*k3[j] + 
			(B4-_B4)*k4[j] + (B5-_B5)*k5[j] + (B6-_B6)*k6[j] + (B7-_B7)*k7[j];
		xNew[j] = x[j] + B1*k1[j] + B2*k2[j] + B3*k3[j] + B4*k4[j] + B5*k5[j] + B6*k6[j] + B7*k7[j];
	}

	// ESTIMATION OF LOCAL ERROR
	rkErr[0] = fabs (rkErr[0]);
	for (j=1; j<NVAR; j++)
		rkErr[0] = rkErr[0] + fabs(rkErr[j]);

	// RETURN ACCEPTANCE (1) OR REJECTION (0) OF step
	if (rkErr[0] < TOL)
		return 1;
	else return 0;
}

#endif

void rk (real_t *x) {
	unsigned char j;
	real_t xNew[NVAR];
	real_t rkErr[NVAR];
	real_t rk2[NVAR];
	real_t step;
	real_t t;
	

	t = 0.0;
	step = 0.001; 
	while (t < TF) {
		if (rkstep (xNew, rkErr, x, step)) {
			t = t + step;
			for (j=0; j<NVAR; j++)
				x[j] = xNew[j];
		}

		// UPDATE stepSIZE FOR NEXT step (OR REPETITION)
		xNew[0]= pow (TOL/rkErr[0], 1.0/ORDERPLUS);			// RATIO FOR NEXT step
		xNew[0] = MAX (xNew[0], 0.1);					// MINIMUM RATIO 1/10
		xNew[0] = MIN (xNew[0], 2.0);					// MAXIMUM RATIO 2
		step = 0.9 * step * xNew[0];	
		step = MIN (step,TF-t);

	}

}

